// license: AGPLv3 or later
// copyright 2023 zamfofex

#include <mimimi/geometry.h>
#include <mimimi/compound-behaviors.h>

#include <mimimi/chapter-I/cameras.h>

static void mimimi_chapter_I_clamp_camera(void *data)
{
	struct mimimi_chapter_I_camera_data *camera_data = data;
	struct mimimi_position *origin = camera_data->origin;
	struct mimimi_position *player_position = camera_data->player_position;
	struct mimimi_position *input = camera_data->input;
	struct mimimi_position *output = camera_data->output;
	
	static struct mimimi_chapter_I_camera_threshold thresholds[] =
	{
		{-36, 0, 0},
		{8, 2, 2},
	};
	
	static int count = sizeof thresholds / sizeof *thresholds;
	
	int i;
	for (i = 1 ; i < count ; i++)
	{
		if (player_position->x / 128 < thresholds[i].x + origin->x)
			break;
	}
	
	int left = (thresholds[i - 1].x + origin->x + 16) * 128;
	int right = (thresholds[i].x + origin->x - 16) * 128;
	
	int top = (thresholds[i].top + origin->y - 4) * 128;
	int bottom = (thresholds[i].bottom + origin->y - 4) * 128;
	
	*output = *input;
	
	if (output->x < left) output->x = left;
	if (output->x > right) output->x = right;
	if (output->y < top) output->y = top;
	if (output->y > bottom) output->y = bottom;
}

void mimimi_chapter_I_camera(struct mimimi_behavior *compound, struct mimimi_chapter_I_camera_data *data, struct mimimi_allocator *allocator)
{
	mimimi_compound(compound, mimimi_function(data, &mimimi_chapter_I_clamp_camera, allocator));
}
