// license: AGPLv3 or later
// copyright 2023 zamfofex

#include <mimimi/chapter-I/cameras.h>
#include <mimimi/sprites.h>
#include <mimimi/controls.h>
#include <mimimi/collisions.h>
#include <mimimi/chapter-I/cameras.h>
#include <mimimi/cameras.h>
#include <mimimi/chapter-I/appearances.h>
#include <mimimi/chapter-I/story.h>
#include <mimimi/chapter-I/models.h>

#include <mimimi/chapter-I.h>

struct mimimi_chapter_I_sprite
{
	int x, y;
	struct mimimi_model *model;
};

static void mimimi_chapter_I_player(struct mimimi_behavior *compound, struct mimimi_sprite *player, struct mimimi_history *history, struct mimimi_allocator *allocator)
{
	struct mimimi_allocator *compound_allocator = mimimi_compound_allocator(compound, allocator);
	
	struct mimimi_jump *jump = mimimi_allocate(compound_allocator, sizeof *jump);
	struct mimimi_behavior *jump_behavior = mimimi_jump(jump, player->physics, compound_allocator);
	
	struct mimimi_behavior *jump_choice = mimimi_choice(&player->physics->airborne, mimimi_noop, jump_behavior, compound_allocator);
	
	struct mimimi_controls *controls = mimimi_allocate(compound_allocator, sizeof *controls);
	mimimi_compound(compound, mimimi_controls(controls, player->walk, history, compound_allocator));
	controls->jump = jump_choice;
}

struct mimimi_chapter *mimimi_chapter_I(struct mimimi_engine *engine)
{
	struct mimimi_behavior *compound = mimimi_compound_behavior(engine->allocator);
	struct mimimi_allocator *compound_allocator = mimimi_compound_allocator(compound, engine->allocator);
	
	struct mimimi_chapter *chapter = mimimi_allocate(compound_allocator, sizeof *chapter);
	chapter->behavior = compound;
	
	struct mimimi_behavior *display = mimimi_compound_behavior(engine->allocator);
	mimimi_compound(compound, display);
	
	struct mimimi_ground *ground = mimimi_allocate(compound_allocator, sizeof *ground);
	ground->width = 64;
	ground->height = 64;
	ground->tiles = mimimi_allocate(compound_allocator, ground->width * ground->height);
	for (int i = 0 ; i < ground->width * ground->height ; i++)
		ground->tiles[i] = 1;
	
	struct mimimi_chapter_I_stage stage = {ground, 0, engine->allocator->null};
	mimimi_chapter_I_hideout(&stage, 1, 1, compound_allocator);
	
	struct mimimi_chapter_I_sprite infos[] =
	{
		{0, 0, mimimi_chapter_I_mirai_flash},
	};
	
	int count = sizeof infos / sizeof *infos;
	
	struct mimimi_position *origin = mimimi_allocate(compound_allocator, sizeof *origin);
	
	for (int y = 0 ; y < ground->height ; y++)
	for (int x = 0 ; x < ground->width ; x++)
	{
		if (ground->tiles[x + y * ground->width] == 2)
		{
			origin->x = x;
			origin->y = y;
			for (int i = 0 ; i < count ; i++)
			{
				infos[i].x += x;
				infos[i].y += y;
			}
		}
	}
	
	struct mimimi_position **positions = mimimi_allocate(compound_allocator, count * sizeof *positions);
	mimimi_compound(compound, mimimi_collision(count, positions, 128, 256, 512, 64, engine->allocator));
	
	struct mimimi_sprite **sprites = mimimi_allocate(engine->allocator, count * sizeof *sprites);
	
	struct mimimi_history *history = mimimi_allocate(compound_allocator, sizeof *history);
	mimimi_compound(compound, mimimi_history(history, chapter, compound_allocator));
	
	struct mimimi_position *camera_input = mimimi_allocate(compound_allocator, sizeof *camera_input);
	struct mimimi_position *camera = mimimi_allocate(compound_allocator, sizeof *camera);
	
	struct mimimi_chapter_I_camera_data *camera_data = mimimi_allocate(compound_allocator, sizeof *camera_data);
	
	camera_data->origin = origin;
	
	for (int i = 0 ; i < count ; i++)
	{
		struct mimimi_chapter_I_sprite *info = infos + i;
		
		struct mimimi_sprite *sprite = mimimi_sprite(ground, info->x * 128, info->y * 128, 80, 250, engine->allocator);
		mimimi_compound(compound, sprite->behavior);
		sprites[i] = sprite;
		
		if (i == 0)
		{
			camera_data->player_position = sprite->position;
			camera_data->input = camera_input;
			camera_data->output = camera;
			
			mimimi_chapter_I_player(compound, sprite, history, engine->allocator);
			mimimi_compound(compound, mimimi_camera(camera_input, sprite->position, 0, -512, engine->allocator));
			mimimi_compound(compound, mimimi_function(camera_data, &mimimi_chapter_I_clamp_camera, engine->allocator));
		}
		else
		{
			// todo
		}
		
		positions[i] = sprite->position;
	}
	
	struct mimimi_sprite *player = sprites[0];
	
	struct mimimi_image *overlay = mimimi_allocate(compound_allocator, sizeof *overlay);
	overlay->width = engine->size->width;
	overlay->height = engine->size->height;
	overlay->colors = mimimi_allocate(compound_allocator, overlay->width * overlay->height);
	
	for (int i = 0 ; i < overlay->width * overlay->height ; i++)
		overlay->colors[i] = 1;
	
	void *overlay_texture = (*engine->texture)(engine->data, overlay);
	mimimi_compound(display, mimimi_overlay(engine, overlay_texture, 0, 0, engine->allocator));
	
	int image_index = 0;
	
	while (image_index < stage.count)
	{
		struct mimimi_chapter_I_image *image = &stage.images[image_index];
		if (image->z < 0) break;
		void *texture = (*engine->texture)(engine->data, image->image);
		mimimi_compound(display, mimimi_background(engine, camera, texture, image->x, image->y, image->z, engine->allocator));
		image_index++;
	}
	
	for (int i = 0 ; i < count ; i++)
	{
		struct mimimi_chapter_I_sprite *info = infos + i;
		struct mimimi_sprite *sprite = sprites[i];
		
		struct mimimi_chapter_I_appearance *appearance = mimimi_chapter_I_appearance(info->model, compound_allocator);
		
		struct mimimi_chapter_I_video_appearance *video_appearance = mimimi_allocate(compound_allocator, sizeof *video_appearance);
		mimimi_chapter_I_video_appearance(video_appearance, appearance, engine, compound_allocator);
		
		mimimi_chapter_I_display(display, engine, camera, sprite, video_appearance, 24, 48, engine->allocator);
	}
	
	mimimi_deallocate(engine->allocator, sprites);
	
	while (image_index < stage.count)
	{
		struct mimimi_chapter_I_image *image = &stage.images[image_index];
		void *texture = (*engine->texture)(engine->data, image->image);
		mimimi_compound(display, mimimi_background(engine, camera, texture, image->x, image->y, image->z + 1, engine->allocator));
		image_index++;
	}
	
	struct mimimi_chapter_I_story *data = mimimi_allocate(compound_allocator, sizeof *data);
	data->chapter = chapter;
	data->engine = engine;
	data->allocator = compound_allocator;
	data->player = player;
	
	mimimi_coroutine(engine->coroutines, compound, &mimimi_chapter_I_story, data, compound_allocator);
	return chapter;
}
