// license: AGPLv3 or later
// copyright 2023 zamfofex

#include <mimimi/chapter-I/stages.h>
#include <mimimi/chapter-I/assets.h>
#include <mimimi/ground.h>

#include <mimimi/chapter-I/stages.h>

void mimimi_chapter_I_add_image(struct mimimi_chapter_I_stage *stage, struct mimimi_image *image, int x, int y, int z, struct mimimi_allocator *allocator)
{
	stage->count++;
	stage->images = mimimi_reallocate(allocator, stage->images, stage->count * sizeof *stage->images);
	stage->images[stage->count - 1].image = image;
	stage->images[stage->count - 1].x = x;
	stage->images[stage->count - 1].y = y;
	stage->images[stage->count - 1].z = z;
	for (int i = stage->count - 1 ; i > 0 ; i--)
	{
		int j = i - 1;
		if (stage->images[i].z < stage->images[j].z) break;
		struct mimimi_chapter_I_image image = stage->images[i];
		stage->images[i] = stage->images[j];
		stage->images[j] = image;
	}
}

static void mimimi_chapter_I_unpack_bits(int size, unsigned char *unpacked, unsigned char *packed)
{
	int i = 0;
	int j = 0;
	
	while (j < size)
	{
		int n = packed[i++];
		if (n == 128) continue;
		if (n > 128) n -= 256;
		
		if (n > 0)
		{
			for (int k = 0 ; k <= n ; k++)
				unpacked[j++] = packed[i++];
		}
		else
		{
			for (int k = 0 ; k <= -n ; k++)
				unpacked[j++] = packed[i];
			i++;
		}
	}
}

void mimimi_chapter_I_hideout(struct mimimi_chapter_I_stage *stage, int x, int y, struct mimimi_allocator *allocator)
{
	int width = 41;
	int height = 11;
	
	static unsigned char tiles[] =
	{
		1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,
		1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,
		1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,
		1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,
		1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,
		1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,
		1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,
		1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,
		1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,
		1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,
		1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,2,1,1,1,1,1,1,
	};
	
	struct mimimi_image *image = mimimi_allocate(allocator, sizeof *image);
	image->width = 699;
	image->height = 256;
	image->colors = mimimi_allocate(allocator, image->width * image->height);
	
	mimimi_chapter_I_unpack_bits(image->width * image->height, image->colors, mimimi_chapter_I_hideout_buffer);
	
	mimimi_chapter_I_add_image(stage, image, ((x - 2) * 16 + 5) * 8, y * 16 * 8, 0, allocator);
	
	for (int y1 = 0 ; y1 < height ; y1++)
	for (int x1 = 0 ; x1 < width ; x1++)
		stage->ground->tiles[(x + x1) + (y + y1) * stage->ground->width] = tiles[x1 + y1 * width];
}
